<?php
// +----------------------------------------------------------------------
// | ThinkPHP [ WE CAN DO IT JUST THINK ]
// +----------------------------------------------------------------------
// | Copyright (c) 2006~2018 http://thinkphp.cn All rights reserved.
// +----------------------------------------------------------------------
// | Licensed ( http://www.apache.org/licenses/LICENSE-2.0 )
// +----------------------------------------------------------------------
// | Author: liu21st <liu21st@gmail.com>
// +----------------------------------------------------------------------

namespace think;

use think\exception\ClassNotFoundException;
use think\exception\HttpResponseException;
use think\route\Dispatch;

/**
 * App 应用管理
 */
class App implements \ArrayAccess
{
    const VERSION = '5.1.5';

    /**
     * 当前模块路径
     * @var string
     */
    protected $modulePath;

    /**
     * 应用调试模式
     * @var bool
     */
    protected $debug = true;

    /**
     * 应用开始时间
     * @var float
     */
    protected $beginTime;

    /**
     * 应用内存初始占用
     * @var integer
     */
    protected $beginMem;

    /**
     * 应用类库命名空间
     * @var string
     */
    protected $namespace = 'app';

    /**
     * 应用类库后缀
     * @var bool
     */
    protected $suffix = false;

    /**
     * 严格路由检测
     * @var bool
     */
    protected $routeMust;

    /**
     * 应用类库目录
     * @var string
     */
    protected $appPath;

    /**
     * 框架目录
     * @var string
     */
    protected $thinkPath;

    /**
     * 应用根目录
     * @var string
     */
    protected $rootPath;

    /**
     * 运行时目录
     * @var string
     */
    protected $runtimePath;

    /**
     * 配置目录
     * @var string
     */
    protected $configPath;

    /**
     * 路由目录
     * @var string
     */
    protected $routePath;

    /**
     * 配置后缀
     * @var string
     */
    protected $configExt;

    /**
     * 应用调度实例
     * @var Dispatch
     */
    protected $dispatch;

    /**
     * 容器对象实例
     * @var Container
     */
    protected $container;

    /**
     * 绑定模块（控制器）
     * @var string
     */
    protected $bind;

    public function __construct($appPath = '')
    {
        $this->appPath   = $appPath ?: realpath(dirname($_SERVER['SCRIPT_FILENAME']) . '/../application') . '/';
        $this->container = Container::getInstance();
    }

    /**
     * 绑定模块或者控制器
     * @access public
     * @param  string $bind
     * @return $this
     */
    public function bind($bind)
    {
        $this->bind = $bind;
        return $this;
    }

    /**
     * 设置应用类库目录
     * @access public
     * @param  string $path 路径
     * @return $this
     */
    public function path($path)
    {
        $this->appPath = $path;
        return $this;
    }

    /**
     * 初始化应用
     * @access public
     * @return void
     */
    public function initialize()
    {
        //返回当前 Unix 时间戳的微秒数
        $this->beginTime   = microtime(true);
        
        //返回分配给 PHP 的内存量
        $this->beginMem    = memory_get_usage();
        
        //函数返回路径中的目录部分 
        // dirname(__DIR__)     D:\wamp\www\tp5\thinkphp\library
        // dirname(dirname(__DIR__))    D:\wamp\www\tp5\thinkphp
        $this->thinkPath   = dirname(dirname(__DIR__)) . '/';       // D:\wamp\www\tp5\thinkphp/
        
        //realpath()    函数返回绝对路径。
        // $this->appPath   D:\wamp\www\tp5\application/
        $this->rootPath    = dirname(realpath($this->appPath)) . '/';   // D:\wamp\www\tp5/
        $this->runtimePath = $this->rootPath . 'runtime/';  // D:\wamp\www\tp5/runtime/
        $this->routePath   = $this->rootPath . 'route/';    // D:\wamp\www\tp5/route/
        $this->configPath  = $this->rootPath . 'config/';   // D:\wamp\www\tp5/config/

        // 设置路径环境变量
        //$this->env    think\Env
        $this->env->set([
            'think_path'   => $this->thinkPath,     // D:\wamp\www\tp5\thinkphp/
            'root_path'    => $this->rootPath,      // D:\wamp\www\tp5/
            'app_path'     => $this->appPath,       // D:\wamp\www\tp5\application/ 应用所在的文件夹
            'config_path'  => $this->configPath,    // D:\wamp\www\tp5/config/
            'route_path'   => $this->routePath,     // D:\wamp\www\tp5/route/
            'runtime_path' => $this->runtimePath,   // D:\wamp\www\tp5/runtime/
            'extend_path'  => $this->rootPath . 'extend/',  // D:\wamp\www\tp5/extend/
            'vendor_path'  => $this->rootPath . 'vendor/',  // D:\wamp\www\tp5/vendor/
        ]);

        // 加载环境变量配置文件
        // is_file() 函数检查指定的文件名是否存在 文件存在则返回 true。
        if (is_file($this->rootPath . '.env')) {
            // 获取.env的配置信息 内容格式为.ini语法
            $this->env->load($this->rootPath . '.env');
        }
        
        $this->namespace = $this->env->get('app_namespace', $this->namespace);  //返回app // $this->namespace
        
        $this->env->set('app_namespace', $this->namespace);

        // 注册应用命名空间 $this->appPath [指定应用所在的文件夹 D:\wamp\www\tp5\application/]
        Loader::addNamespace($this->namespace, $this->appPath);
        
        // 配置后缀
        $this->configExt = $this->env->get('config_ext', '.php');

        // 初始化应用
        $this->init();

        // 开启类名后缀 实际是查看config 内 app内的 class_suffix 值，默认为false
        $this->suffix = $this->config('app.class_suffix');

        // 应用调试模式
        // 如果app_debug 没有设置 则启用 $this->config('app.app_debug')的值 默认为false
        $this->debug = $this->env->get('app_debug', $this->config('app.app_debug'));
        //获取的值设置为 app_debug 的值给予下次调用
        $this->env->set('app_debug', $this->debug);
        
        if (!$this->debug) {
            // 非调试模式 关闭（Off）错误
            ini_set('display_errors', 'Off');
        } elseif (PHP_SAPI != 'cli') {
            //开启调试后切非cli格式 执行
            //重新申请一块比较大的buffer
            if (ob_get_level() > 0) {
                $output = ob_get_clean();
            }
            ob_start();
            if (!empty($output)) {
                echo $output;
            }
        }

        // 注册根命名空间
        // root_namespace 默认为 空数组 
        if (!empty($this->config('app.root_namespace'))) {
            //等于非空数组时执行
            Loader::addNamespace($this->config('app.root_namespace'));
        }

        // 注册类库别名
        // alias 主页为空
        Loader::addClassAlias($this->config->pull('alias'));

        // 设置系统时区
        date_default_timezone_set($this->config('app.default_timezone'));

        // 监听app_init
        $this->hook->listen('app_init',['ceshi','TESTT']);
    }

    /**
     * 初始化应用或模块
     * @access public
     * @param  string $module 模块名
     * @return void
     */
    public function init($module = '')
    {
        // 定位模块目录文件名，指定当前模块的文件夹名称
        $module = $module ? $module . DIRECTORY_SEPARATOR : '';

        //指定当前模块的文件夹目录
        $path   = $this->appPath . $module;
        
        // 加载初始化文件
        // 检测init.php 是否存在
        if (is_file($path . 'init.php')) {
            include $path . 'init.php';
            //检测根目录下的是否存在 D:\wamp\www\tp5/runtime/
        } elseif (is_file($this->runtimePath . $module . 'init.php')) {
            include $this->runtimePath . $module . 'init.php';
        } else {
            // 加载行为扩展文件
            if (is_file($path . 'tags.php')) {
                // $this->hook  think\Hook
                // 注册钩子
                $this->hook->import(include $path . 'tags.php');
            }

            // 加载application下的 common.php 公共文件
            if (is_file($path . 'common.php')) {
                include $path . 'common.php';
            }
            
            //$module 等于空 加载系统助手函数
            if ('' == $module) {
                // 加载系统助手函数 D:\wamp\www\tp5\thinkphp/
                include $this->thinkPath . 'helper.php';
            }

            // 注册服务的容器对象实例
            if (is_file($path . 'provider.php')) {
                $this->container->bind(include $path . 'provider.php');
            }

            // 自动读取配置文件
            //模块下的配置文件夹不存在就加载 D:\wamp\www\tp5/config/ 内的
            if (is_dir($path . 'config')) {
                $dir = $path . 'config';
            } elseif (is_dir($this->configPath . $module)) {
                $dir = $this->configPath . $module;
            }
            
            // scandir($dir) 列出 $dir 目录中的文件和目录
            $files = isset($dir) ? scandir($dir) : [];
            
            //循环去掉一切不需要的值
            foreach ($files as $file) {
                if ('.' . pathinfo($file, PATHINFO_EXTENSION) === $this->configExt) {
                    $filename = $dir . DIRECTORY_SEPARATOR . $file;
                    //加载所有的配置文件
                    // $this->config think\Config
                    $this->config->load($filename, pathinfo($file, PATHINFO_FILENAME));
                }
            }
        }

        // $this->request  think\Request
        // 全局过滤方法 default_filter 实际是 app 键值内的 default_filter 值
        $this->request->filter($this->config('app.default_filter'));
    }

    /**
     * 执行应用程序
     * @access public
     * @return Response
     * @throws Exception
     */
    public function run()
    {
        // 初始化应用,初步完成注释
        $this->initialize();
        
        try {
            if ($this->bind) {
                // 模块/控制器绑定
                // $this->route     think\Route
                $this->route->bind($this->bind);
            } elseif ($this->config('app.auto_bind_module')) {
                // 入口自动绑定模块 app.auto_bind_module 默认是false
                // 入口自动绑定
                $name = pathinfo($this->request->baseFile(), PATHINFO_FILENAME);
                if ($name && 'index' != $name && is_dir($this->appPath . $name)) {
                    $this->route->bind($name);
                }
            }

            // 读取默认语言
            // app.default_lang 默认为 zh-cn
            // $this->lang      think\Lang
            // 设置当前的语言
            $this->lang->range($this->config('app.default_lang'));
            //判断是否开启了多语言，默认是false
            if ($this->config('app.lang_switch_on')) {
                // 开启多语言机制 检测当前语言
                $this->lang->detect();
            }

            // $this->request  think\Request
            // $this->lang->range() 获取当前的语言
            // 设置当前的语言
            $this->request->langset($this->lang->range());

            // 加载系统语言包
            $this->lang->load([
                $this->thinkPath . 'lang/' . $this->request->langset() . '.php',
                $this->appPath . 'lang/' . $this->request->langset() . '.php',
            ]);

            // 监听app_dispatch
            $this->hook->listen('app_dispatch');

            // 获取应用调度信息
            $dispatch = $this->dispatch;
            if (empty($dispatch)) {
                //$dispatch 为空时执行
                // 进行URL路由检测
                // 主页返回的 think\route\dispatch\Url 实例
                // 在这里也或许会返回404
                $dispatch = $this->routeCheck();
            }

            // 记录当前调度信息 设置或者获取当前请求的调度信息
            $this->request->dispatch($dispatch);

            // 记录路由和请求信息
            if ($this->debug) {
                $this->log('[ ROUTE ] ' . var_export($this->request->routeInfo(), true));
                $this->log('[ HEADER ] ' . var_export($this->request->header(), true));
                $this->log('[ PARAM ] ' . var_export($this->request->param(), true));
            }

            // 监听app_begin
            $this->hook->listen('app_begin');

            // 请求缓存检查
            $this->request->cache(
                $this->config('app.request_cache'),
                $this->config('app.request_cache_expire'),
                $this->config('app.request_cache_except')
            );

            // 执行调度
            // think\route\dispatch\Url 内的 run()
            $data = $dispatch->run();
            var_dump($data);   

//  EXIT;

        } catch (HttpResponseException $exception) {
            $data = $exception->getResponse();
        }

        $this->middlewareDispatcher->add(function (Request $request, $next) use ($data) {
            // 输出数据到客户端
            if ($data instanceof Response) {
                $response = $data;
            } elseif (!is_null($data)) {
                // 默认自动识别响应输出类型
                $isAjax = $request->isAjax();
                $type   = $isAjax ? $this->config('app.default_ajax_return') : $this->config('app.default_return_type');

                $response = Response::create($data, $type);
            } else {
                $response = Response::create();
            }
            return $response;
        });

        $response = $this->middlewareDispatcher->dispatch($this->request);

        // 监听app_end
        $this->hook->listen('app_end', $response);

        return $response;
    }

    /**
     * 设置当前请求的调度信息
     * @access public
     * @param  Dispatch  $dispatch 调度信息
     * @return $this
     */
    public function dispatch(Dispatch $dispatch)
    {
        $this->dispatch = $dispatch;
        return $this;
    }

    /**
     * 记录调试信息
     * @access public
     * @param  mixed  $msg  调试信息
     * @param  string $type 信息类型
     * @return void
     */
    public function log($log, $type = 'info')
    {
        $this->debug && $this->log->record($log, $type);
    }

    /**
     * 获取配置参数 为空则获取所有配置
     * @access public
     * @param  string    $name 配置参数名（支持二级配置 .号分割）
     * @return mixed
     */
    public function config($name = '')
    {
        return $this->config->get($name);
    }

    /**
     * URL路由检测（根据PATH_INFO)
     * @access public
     * @return Dispatch
     */
    public function routeCheck()
    {
        // 获取当前请求URL的pathinfo信息(不含URL后缀 如：.html)
        $path = $this->request->path();
        // pathinfo_depr 为 pathinfo分隔符，默认为'/'
        $depr = $this->config('app.pathinfo_depr');

        // 路由检测  
        $files = scandir($this->routePath);     // D:\wamp\www\tp5/route/
        foreach ($files as $file) {
            if (strpos($file, '.php')) {
                $filename = $this->routePath . $file;
                var_dump($filename);
                // 导入路由配置
                $rules = include $filename;
                // is_array() 检测变量是否是数组
                if (is_array($rules)) {
                    $this->route->import($rules);
                    
                }
            }
        }
        $this->route->routeinfo();
//        exit;

        // app.route_annotation 使用注解路由 默认false
        // 这个模式可以放弃使用
        if ($this->config('app.route_annotation')) {
            // 自动生成路由定义
            if ($this->debug) {
                $this->build->buildRoute($this->config('app.controller_suffix'));
            }

            $filename = $this->runtimePath . 'build_route.php';

            if (is_file($filename)) {
                include $filename;
            }
        }
        
        // 是否强制路由模式
        // 等于false 则使用
        // $this->routeMust 默认为空 is_null($this->routeMust) 为 true
        // $this->routeMust 默认为空
        // app.url_route_must 是否强制使用路由 默认为false
        $must = !is_null($this->routeMust) ? $this->routeMust : $this->config('app.url_route_must');
        // $must    值为false

        // 路由检测 返回一个Dispatch对象
        
        // $path 请求URL 不含后缀(.HTML) 不含域名
        // URL分隔符 默认为/
        // $must  是否强制路由模式 默认值为false
        // app.rote_complete_match 路由是否完全匹配 默认为false
        return $this->route->check($path, $depr, $must, $this->config('app.route_complete_match'));
    }

    /**
     * 设置应用的路由检测机制
     * @access public
     * @param  bool $must  是否强制检测路由
     * @return $this
     */
    public function routeMust($must = false)
    {
        $this->routeMust = $must;
        return $this;
    }

    /**
     * 解析模块和类名
     * @access protected
     * @param  string $name         资源地址
     * @param  string $layer        验证层名称
     * @param  bool   $appendSuffix 是否添加类名后缀
     * @return array
     */
    protected function parseModuleAndClass($name, $layer, $appendSuffix)
    {
        if (false !== strpos($name, '\\')) {
            $class  = $name;
            $module = $this->request->module();
        } else {
            if (strpos($name, '/')) {
                list($module, $name) = explode('/', $name, 2);
            } else {
                $module = $this->request->module();
            }

            $class = $this->parseClass($module, $layer, $name, $appendSuffix);
        }

        return [$module, $class];
    }

    /**
     * 实例化应用类库
     * @access public
     * @param  string $name         类名称
     * @param  string $layer        业务层名称
     * @param  bool   $appendSuffix 是否添加类名后缀
     * @param  string $common       公共模块名
     * @return object
     * @throws ClassNotFoundException
     */
    public function create($name, $layer, $appendSuffix = false, $common = 'common')
    {
        $guid = $name . $layer;

        if ($this->__isset($guid)) {
            return $this->__get($guid);
        }

        list($module, $class) = $this->parseModuleAndClass($name, $layer, $appendSuffix);

        if (class_exists($class)) {
            $object = $this->__get($class);
        } else {
            $class = str_replace('\\' . $module . '\\', '\\' . $common . '\\', $class);
            if (class_exists($class)) {
                $object = $this->__get($class);
            } else {
                throw new ClassNotFoundException('class not exists:' . $class, $class);
            }
        }

        $this->__set($guid, $class);

        return $object;
    }

    /**
     * 实例化（分层）模型
     * @access public
     * @param  string $name         Model名称
     * @param  string $layer        业务层名称
     * @param  bool   $appendSuffix 是否添加类名后缀
     * @param  string $common       公共模块名
     * @return Model
     * @throws ClassNotFoundException
     */
    public function model($name = '', $layer = 'model', $appendSuffix = false, $common = 'common')
    {
        return $this->create($name, $layer, $appendSuffix, $common);
    }

    /**
     * 实例化（分层）控制器 格式：[模块名/]控制器名
     * @access public
     * @param  string $name              资源地址
     * @param  string $layer             控制层名称
     * @param  bool   $appendSuffix      是否添加类名后缀
     * @param  string $empty             空控制器名称
     * @return object
     * @throws ClassNotFoundException
     */
    public function controller($name, $layer = 'controller', $appendSuffix = false, $empty = '')
    {
        list($module, $class) = $this->parseModuleAndClass($name, $layer, $appendSuffix);

        if (class_exists($class)) {
            return $this->__get($class);
        } elseif ($empty && class_exists($emptyClass = $this->parseClass($module, $layer, $empty, $appendSuffix))) {
            return $this->__get($emptyClass);
        } else {
            throw new ClassNotFoundException('class not exists:' . $class, $class);
        }
    }

    /**
     * 实例化验证类 格式：[模块名/]验证器名
     * @access public
     * @param  string $name         资源地址
     * @param  string $layer        验证层名称
     * @param  bool   $appendSuffix 是否添加类名后缀
     * @param  string $common       公共模块名
     * @return Validate
     * @throws ClassNotFoundException
     */
    public function validate($name = '', $layer = 'validate', $appendSuffix = false, $common = 'common')
    {
        $name = $name ?: $this->config('default_validate');

        if (empty($name)) {
            return new Validate;
        }

        return $this->create($name, $layer, $appendSuffix, $common);
    }

    /**
     * 数据库初始化
     * @access public
     * @param  mixed         $config 数据库配置
     * @param  bool|string   $name 连接标识 true 强制重新连接
     * @return \think\db\Query
     */
    public function db($config = [], $name = false)
    {
        return Db::connect($config, $name);
    }

    /**
     * 远程调用模块的操作方法 参数格式 [模块/控制器/]操作
     * @access public
     * @param  string       $url          调用地址
     * @param  string|array $vars         调用参数 支持字符串和数组
     * @param  string       $layer        要调用的控制层名称
     * @param  bool         $appendSuffix 是否添加类名后缀
     * @return mixed
     * @throws ClassNotFoundException
     */
    public function action($url, $vars = [], $layer = 'controller', $appendSuffix = false)
    {
        $info   = pathinfo($url);
        $action = $info['basename'];
        $module = '.' != $info['dirname'] ? $info['dirname'] : $this->request->controller();
        $class  = $this->controller($module, $layer, $appendSuffix);

        if (is_scalar($vars)) {
            if (strpos($vars, '=')) {
                parse_str($vars, $vars);
            } else {
                $vars = [$vars];
            }
        }

        return $this->container->invokeMethod([$class, $action . $this->config('action_suffix')], $vars);
    }

    /**
     * 解析应用类的类名
     * @access public
     * @param  string $module 模块名
     * @param  string $layer  层名 controller model ...
     * @param  string $name   类名
     * @param  bool   $appendSuffix
     * @return string
     */
    public function parseClass($module, $layer, $name, $appendSuffix = false)
    {
        $name  = str_replace(['/', '.'], '\\', $name);
        $array = explode('\\', $name);
        $class = Loader::parseName(array_pop($array), 1) . ($this->suffix || $appendSuffix ? ucfirst($layer) : '');
        $path  = $array ? implode('\\', $array) . '\\' : '';

        return $this->namespace . '\\' . ($module ? $module . '\\' : '') . $layer . '\\' . $path . $class;
    }

    /**
     * 获取框架版本
     * @access public
     * @return string
     */
    public function version()
    {
        return static::VERSION;
    }

    /**
     * 是否为调试模式
     * @access public
     * @return bool
     */
    public function isDebug()
    {
        return $this->debug;
    }

    /**
     * 获取模块路径
     * @access public
     * @return string
     */
    public function getModulePath()
    {
        return $this->modulePath;
    }

    /**
     * 设置模块路径
     * @access public
     * @param  string $path 路径
     * @return void
     */
    public function setModulePath($path)
    {
        $this->modulePath = $path;
        $this->env->set('module_path', $path);
    }

    /**
     * 获取应用根目录
     * @access public
     * @return string
     */
    public function getRootPath()
    {
        return $this->rootPath;
    }

    /**
     * 获取应用类库目录
     * @access public
     * @return string
     */
    public function getAppPath()
    {
        return $this->appPath;
    }

    /**
     * 获取应用运行时目录
     * @access public
     * @return string
     */
    public function getRuntimePath()
    {
        return $this->runtimePath;
    }

    /**
     * 获取核心框架目录
     * @access public
     * @return string
     */
    public function getThinkPath()
    {
        return $this->thinkPath;
    }

    /**
     * 获取路由目录
     * @access public
     * @return string
     */
    public function getRoutePath()
    {
        return $this->routePath;
    }

    /**
     * 获取应用配置目录
     * @access public
     * @return string
     */
    public function getConfigPath()
    {
        return $this->configPath;
    }

    /**
     * 获取配置后缀
     * @access public
     * @return string
     */
    public function getConfigExt()
    {
        return $this->configExt;
    }

    /**
     * 获取应用类库命名空间
     * @access public
     * @return string
     */
    public function getNamespace()
    {
        return $this->namespace;
    }

    /**
     * 设置应用类库命名空间
     * @access public
     * @param  string $namespace 命名空间名称
     * @return $this
     */
    public function setNamespace($namespace)
    {
        $this->namespace = $namespace;
        return $this;
    }

    /**
     * 是否启用类库后缀
     * @access public
     * @return bool
     */
    public function getSuffix()
    {
        return $this->suffix;
    }

    /**
     * 获取应用开启时间
     * @access public
     * @return float
     */
    public function getBeginTime()
    {
        return $this->beginTime;
    }

    /**
     * 获取应用初始内存占用
     * @access public
     * @return integer
     */
    public function getBeginMem()
    {
        return $this->beginMem;
    }

    /**
     * 获取容器实例
     * @access public
     * @return Container
     */
    public function container()
    {
        return $this->container;
    }

    public function __set($name, $value)
    {
        $this->container->bind($name, $value);
    }

    public function __get($name)
    {
        return $this->container->make($name);
    }

    public function __isset($name)
    {
        return $this->container->bound($name);
    }

    public function __unset($name)
    {
        $this->container->__unset($name);
    }

    public function offsetExists($key)
    {
        return $this->__isset($key);
    }

    public function offsetGet($key)
    {
        return $this->__get($key);
    }

    public function offsetSet($key, $value)
    {
        $this->__set($key, $value);
    }

    public function offsetUnset($key)
    {
        $this->__unset($key);
    }
}
