<?php
// +----------------------------------------------------------------------
// | ThinkPHP [ WE CAN DO IT JUST THINK ]
// +----------------------------------------------------------------------
// | Copyright (c) 2006~2018 http://thinkphp.cn All rights reserved.
// +----------------------------------------------------------------------
// | Licensed ( http://www.apache.org/licenses/LICENSE-2.0 )
// +----------------------------------------------------------------------
// | Author: liu21st <liu21st@gmail.com>
// +----------------------------------------------------------------------

namespace think;

class Config implements \ArrayAccess
{
    /**
     * 配置参数
     * @var array
     */
    private $config = [];

    /**
     * 缓存前缀
     * @var string
     */
    private $prefix = 'app';

    /**
     * 设置配置参数默认前缀
     * @access public
     * @param string    $prefix 前缀
     * @return void
     */
    public function setDefaultPrefix($prefix)
    {
        $this->prefix = $prefix;
    }

    /**
     * 解析配置文件或内容
     * @access public
     * @param  string    $config 配置文件路径或内容
     * @param  string    $type 配置解析类型
     * @param  string    $name 配置名（如设置即表示二级配置）
     * @return mixed
     */
    public function parse($config, $type = '', $name = '')
    {
        if (empty($type)) {
            $type = pathinfo($config, PATHINFO_EXTENSION);
        }

        $class = false !== strpos($type, '\\') ? $type : '\\think\\config\\driver\\' . ucwords($type);

        return $this->set((new $class())->parse($config), $name);
    }

    /**
     * 加载配置文件（多种格式）
     * @access public
     * @param  string    $file 配置文件名
     * @param  string    $name 一级配置名
     * @return mixed
     */
    public function load($file, $name = '')
    {
        //检测文件是否存在
        if (is_file($file)) {
            //$name 所有字符转换为小写
            $name = strtolower($name);
            //获取文件的后缀格式
            $type = pathinfo($file, PATHINFO_EXTENSION);
            if ('php' == $type) {
                // 后缀为 php格式，返回配置信息数组的值
                return $this->set(include $file, $name);
            } elseif ('yaml' == $type && function_exists('yaml_parse_file')) {
                // yaml 后缀格式 并检测环境是否支持 yaml_parse_file 函数
                return $this->set(yaml_parse_file($file), $name);
            } else {
                // 除以上两种格式以外的
                return $this->parse($file, $type, $name);
            }
        } else {
            // 文件不存在直接返回配置信息数组
            return $this->config;
        }
    }

    /**
     * 自动加载配置文件（PHP格式）
     * @access public
     * @param  string    $name 配置名
     * @return void
     */
    protected function autoLoad($name)
    {
        // 如果尚未载入 则动态加载配置文件
        $module = Container::get('request')->module();
        $module = $module ? $module . '/' : '';
        $app    = Container::get('app');
        $path   = $app->getAppPath() . $module;

        if (is_dir($path . 'config')) {
            $file = $path . 'config/' . $name . $app->getConfigExt();
        } elseif (is_dir($app->getConfigPath() . $module)) {
            $file = $app->getConfigPath() . $module . $name . $app->getConfigExt();
        }

        if (isset($file) && is_file($file)) {
            $this->load($file, $name);
        }
    }

    /**
     * 检测配置是否存在
     * @access public
     * @param  string    $name 配置参数名（支持多级配置 .号分割）
     * @return bool
     */
    public function has($name)
    {
        if (!strpos($name, '.')) {
            $name = $this->prefix . '.' . $name;
        }

        return $this->get($name) ? true : false;
    }

    /**
     * 获取一级配置
     * @access public
     * @param  string    $name 一级配置名
     * @return array
     */
    public function pull($name)
    {
        $name = strtolower($name);

        if (!isset($this->config[$name])) {
            // 如果尚未载入 则动态加载配置文件
            $this->autoLoad($name);
        }

        return isset($this->config[$name]) ? $this->config[$name] : [];
    }

    /**
     * 获取配置参数 为空则获取所有配置
     * @access public
     * @param  string    $name 配置参数名（支持多级配置 .号分割）
     * @return mixed
     */
    public function get($name = null)
    {
        // 无参数时获取所有
        if (empty($name)) {
            return $this->config;
        }
        
        //点在字符串中第一次出现的位置
        if (!strpos($name, '.')) {
            $name = $this->prefix . '.' . $name;
        } elseif ('.' == substr($name, -1)) {
            return $this->pull(substr($name, 0, -1));
        }

        $name    = explode('.', $name);
        $name[0] = strtolower($name[0]);
        $config  = $this->config;

        if (!isset($config[$name[0]])) {
            // 如果尚未载入 则动态加载配置文件
            $this->autoLoad($name[0]);
        }
        
        // 按.拆分成多维数组进行判断
        foreach ($name as $val) {
            if (isset($config[$val])) {
                $config = $config[$val];
            } else {
                return;
            }
        }

        return $config;
    }

    /**
     * 设置配置参数 name为数组则为批量设置
     * @access public
     * @param  string|array  $name 配置参数名（支持三级配置 .号分割）
     * @param  mixed         $value 配置值
     * @return mixed
     */
    public function set($name, $value = null)
    {
        
        // PHP 后缀时 $value 标识为文件名 $name 文件内容（数组）
        // 检测$name 是否是字符串 是为true
        // php 时一般$name 为数组 非字符串
        if (is_string($name)) {
            if (!strpos($name, '.')) {
                $name = $this->prefix . '.' . $name;
            }

            $name = explode('.', $name, 3);

            if (count($name) == 2) {
                $this->config[strtolower($name[0])][$name[1]] = $value;
            } else {
                $this->config[strtolower($name[0])][$name[1]][$name[2]] = $value;
            }

            return $value;
        } elseif (is_array($name)) {
            // 为数组时执行 批量设置
            // 检测￥value 是否为空或者null 不为空才执行如下
            if (!empty($value)) {
                //检测 是否已经存在
                if (isset($this->config[$value])) {
                    //存在则合并成新的
                    $result = array_merge($this->config[$value], $name);
                } else {
                    //不存在则赋值
                    $result = $name;
                }
                //最终根据￥value值为键值
                $this->config[$value] = $result;
            } else {
                // empty($value) 为空时执行 数组合并
                $result = $this->config = array_merge($this->config, $name);
            }
        } else {
            // 为空直接返回 已有配置
            $result = $this->config;
        }
        //并且返回当前的值
        return $result;
    }

    /**
     * 移除配置
     * @access public
     * @param  string  $name 配置参数名（支持三级配置 .号分割）
     * @return void
     */
    public function remove($name)
    {
        if (!strpos($name, '.')) {
            $name = $this->prefix . '.' . $name;
        }

        $name = explode('.', $name, 3);

        if (count($name) == 2) {
            unset($this->config[strtolower($name[0])][$name[1]]);
        } else {
            unset($this->config[strtolower($name[0])][$name[1]][$name[2]]);
        }
    }

    /**
     * 重置配置参数
     * @access public
     * @param  string    $prefix  配置前缀名
     * @return void
     */
    public function reset($prefix = '')
    {
        if ('' === $prefix) {
            $this->config = [];
        } else {
            $this->config[$prefix] = [];
        }
    }

    /**
     * 设置配置
     * @access public
     * @param  string    $name  参数名
     * @param  mixed     $value 值
     */
    public function __set($name, $value)
    {
        return $this->set($name, $value);
    }

    /**
     * 获取配置参数
     * @access public
     * @param  string $name 参数名
     * @return mixed
     */
    public function __get($name)
    {
        return $this->get($name);
    }

    /**
     * 检测是否存在参数
     * @access public
     * @param  string $name 参数名
     * @return bool
     */
    public function __isset($name)
    {
        return $this->has($name);
    }

    // ArrayAccess
    public function offsetSet($name, $value)
    {
        $this->set($name, $value);
    }

    public function offsetExists($name)
    {
        return $this->has($name);
    }

    public function offsetUnset($name)
    {
        $this->remove($name);
    }

    public function offsetGet($name)
    {
        return $this->get($name);
    }
}
