<?php
// +----------------------------------------------------------------------
// | ThinkPHP [ WE CAN DO IT JUST THINK ]
// +----------------------------------------------------------------------
// | Copyright (c) 2006~2018 http://thinkphp.cn All rights reserved.
// +----------------------------------------------------------------------
// | Licensed ( http://www.apache.org/licenses/LICENSE-2.0 )
// +----------------------------------------------------------------------
// | Author: liu21st <liu21st@gmail.com>
// +----------------------------------------------------------------------

namespace think;

class Hook
{
    /**
     * 钩子行为定义
     * @var array
     */
    private $tags = [];

    /**
     * 绑定行为列表
     * @var array
     */
    protected $bind = [];

    /**
     * 入口方法名称
     * @var string
     */
    private static $portal = 'run';

    /**
     * 指定入口方法名称
     * @access public
     * @param  string  $name     方法名
     * @return $this
     */
    public function portal($name)
    {
        self::$portal = $name;
        return $this;
    }

    /**
     * 指定行为标识 便于调用
     * @access public
     * @param  string|array  $name     行为标识
     * @param  mixed         $behavior 行为
     * @return $this
     */
    public function alias($name, $behavior = null)
    {
        if (is_array($name)) {
            $this->bind = array_merge($this->bind, $name);
        } else {
            $this->bind[$name] = $behavior;
        }

        return $this;
    }

    /**
     * 动态添加行为扩展到某个标签
     * @access public
     * @param  string    $tag 标签名称
     * @param  mixed     $behavior 行为名称
     * @param  bool      $first 是否放到开头执行
     * @return void
     */
    public function add($tag, $behavior, $first = false)
    {
        // isset($this->tags[$tag]) 为空 后面执行 $this->tags[$tag] = [];
        // 值为空或者不存在则赋值为数组
        isset($this->tags[$tag]) || $this->tags[$tag] = [];
        // is_callable 检测参数是否为合法的可调用结构 可调用则返回 TRUE
        // $behavior 是数组且不可调用时执行
        if (is_array($behavior) && !is_callable($behavior)) {
            // 检查键名内是否包含 _overlay 不包含时执行
            if (!array_key_exists('_overlay', $behavior)) {
                //执行合并
                $this->tags[$tag] = array_merge($this->tags[$tag], $behavior);
            } else {
                //包含时执行注销
                unset($behavior['_overlay']);
                $this->tags[$tag] = $behavior;
            }
        //开头执行
        } elseif ($first) {
            //$behavior 插入到 $this->tags 数组内
            array_unshift($this->tags[$tag], $behavior);
        } else {
            $this->tags[$tag][] = $behavior;
        }
    }

    /**
     * 批量导入插件
     * @access public
     * @param  array     $tags 插件信息
     * @param  bool      $recursive 是否递归合并
     * @return void
     */
    public function import(array $tags, $recursive = true)
    {
        if ($recursive) {
            foreach ($tags as $tag => $behavior) {
                //将插件信息添加
                $this->add($tag, $behavior);
            }
        } else {
            //数组合并
            $this->tags = $tags + $this->tags;
        }
        
    }

    /**
     * 获取插件信息
     * @access public
     * @param  string $tag 插件位置 留空获取全部
     * @return array
     */
    public function get($tag = '')
    {
        if (empty($tag)) {
            //获取全部的插件信息
            return $this->tags;
        } else {
            return array_key_exists($tag, $this->tags) ? $this->tags[$tag] : [];
        }
    }

    /**
     * 监听标签的行为
     * @access public
     * @param  string $tag    标签名称
     * @param  mixed  $params 传入参数
     * @param  bool   $once   只获取一个有效返回值
     * @return mixed
     */
    public function listen($tag, $params = null, $once = false)
    {
        // $results 用来存放返回的值
        $results = [];
        //获取这个钩子需要执行的所有代码
        $tags    = $this->get($tag);
        
        //若数组内有值，则循环
        foreach ($tags as $key => $name) {
            //根据值执行代码并获取返回值
            // $name 可理解为 需要执行的代码，或者类空间名，或者理解为执行方法
            // $tag 可理解为钩子名称
            // $params 需要传递的参数
            $results[$key] = $this->execTag($name, $tag, $params);
            if (false === $results[$key]) {
                // 如果返回false 则中断行为执行
                break;
            } elseif (!is_null($results[$key]) && $once) {
                //返回值 是 null 则返回 TRUE，
                //返回值 不为空，则只要一个有效值时跳出
                break;
            }
        }

        return $once ? end($results) : $results;
    }

    /**
     * 执行行为
     * @access public
     * @param  mixed     $class  行为
     * @param  mixed     $params 参数
     * @return mixed
     */
    public function exec($class, $params = null)
    {
        if ($class instanceof \Closure || is_array($class)) {
            $method = $class;
        } else {
            if (isset($this->bind[$class])) {
                $class = $this->bind[$class];
            }
            $method = [$class, self::$portal];
        }

        return Container::getInstance()->invoke($method, [$params]);
    }

    /**
     * 执行某个标签的行为
     * @access protected
     * @param  mixed     $class  要执行的行为
     * @param  string    $tag    方法名（标签名）
     * @param  mixed     $params 参数
     * @return mixed
     */
    protected function execTag($class, $tag = '', $params = null)
    {
        //返回app的实例化
        $app = Container::get('app');
        //是否为调试模式，开启调试模式则进行赋值
        $app->isDebug() && $app['debug']->remark('behavior_start', 'time');

        // 字符串命名风格转换
        // 0 将Java风格转换为C的风格 1 将C风格转换为Java的风格,去掉下划线
        // $tag 可理解为钩子名称
        $method = Loader::parseName($tag, 1, false);

        // instanceof 判断一个对象是否是某个类的实例
        if ($class instanceof \Closure) {
            // 闭包调用
            $call  = $class;
            $class = 'Closure'; //调试才使用此值
        } elseif (strpos($class, '::')) {
            // 静态调用
            // strpos 在字符串中第一次出现的位置
            $call = $class;
        } else {
            // 动态调用
            //根据$class 获取此类的实例化
            $obj = Container::get($class);  //会导致先初始化一次，应该缓存一下
            
            //is_callable 检测回调函数 可调用则返回 TRUE
            if (!is_callable([$obj, $method])) {
                // 不可回调 赋默认值
                $method = self::$portal;    // 默认值run() 方法
            }

            $call  = [$class, $method];     // 类或者空间名称  执行方法
            $class = $class . '->' . $method;   //只针对调试使用，拼接成 空间名称执行的例子

        }

        //以上代码主要获取 $call--正常要的参数 和 $class--调试时才需要的参数
        // Container::getInstance() 获取的实例 think\Container，执行 container  内的 invoke()方法
        $result = Container::getInstance()->invoke($call, [$params]);

        //调试模式
        if ($app->isDebug()) {
            $debug = $app['debug'];
            $debug->remark('behavior_end', 'time');
            $app->log('[ BEHAVIOR ] Run ' . $class . ' @' . $tag . ' [ RunTime:' . $debug->getRangeTime('behavior_start', 'behavior_end') . 's ]');
        }

        return $result;
    }

}
